from agox.models.local_GPR.LSGPR_CUR import LSGPRModelCUR
from agox.models.descriptors.soap import SOAP
from agox.models.priors.repulsive import Repulsive
from sklearn.gaussian_process.kernels import RBF, ConstantKernel as C

from agox.models.datasets import datasets

data = datasets["Ag5O3"]
training_data = data[:80]

descriptor = SOAP(["O", "Ag"], r_cut=5.0, nmax=3, lmax=2, sigma=1.0, weight=True, periodic=True)
kernel = C(1) * RBF(length_scale=20)
model = LSGPRModelCUR(
    kernel=kernel, descriptor=descriptor, noise=0.01, prior=Repulsive(ratio=0.7), m_points=1000,
)

model.train_model(training_data)

model.save("my-model")



import os
from argparse import ArgumentParser

from ase.io import read, write

from agox.databases.database import Database
from agox.models.descriptors.voronoi import Voronoi


class Analysis:
    def __init__(self, descriptor, directories, sample_size, save_data=True, force_reload=False):
        self.descriptor = descriptor
        self.directories = directories
        self.sample_size = sample_size
        self.analysed_directories = []

        self.save_data = save_data
        self.force_reload = force_reload

        self.structures = []

    def analyse_directories(self):
        for idx, directory in enumerate(self.directories):
            self.add_directory(directory, force_reload=self.force_reload)

    def add_directory(self, directory, force_reload=False):
        if directory in self.analysed_directories:
            print("Directory: {} - already been loaded so skipping it".format(directory))

        cwd = os.getcwd()
        full_path = os.path.join(cwd, directory)
        if full_path[-1] != "/":
            full_path += "/"
        data_path = full_path + "GS_data/"

        if os.path.exists(data_path) and not force_reload:
            print("{} has already been analysed.".format(full_path))
            self.structures += read(data_path + "unique.traj", index=":")
            if self.descriptor.template is None:
                template = read(data_path + "template.traj", index="0")
                self.descriptor.template = template

        else:
            # Find database files in this directory:
            structures = []
            print("Loading files from directory: {}".format(directory))
            for f in os.listdir(os.path.join(cwd, directory)):
                if f[-3:] == ".db":
                    db = Database(os.path.join(cwd, directory, f))
                    db.restore_to_memory()
                    structures += db.get_all_candidates()
                    if self.descriptor.template is None:
                        template = structures[0].get_template()
                        self.descriptor.template = template

            unique = self.sort_structures(structures)
            self.structures += unique

            if self.save_data:
                if not os.path.exists(data_path):
                    os.mkdir(data_path)
                write(data_path + "unique.traj", unique)
                write(data_path + "template.traj", self.descriptor.template)

        self.analysed_directories.append(directory)

    def sort_structures(self, structures):
        structures = sorted(structures, key=lambda x: x.get_potential_energy())
        print(f"Total number of structures: {len(structures)}")
        unique = []
        eigen_spectrum = []
        for i, structure in enumerate(structures):
            eigen_values = self.descriptor.create_features(structure)
            if eigen_values not in eigen_spectrum:
                unique.append(structure)
                eigen_spectrum.append(eigen_values)
                print(i, eigen_values)
                if len(unique) == self.sample_size:
                    break
        print(f"Number of unique structures: {len(unique)}")
        return unique


def analysis():
    parser = ArgumentParser()
    parser.add_argument("-d", "--directories", nargs="+", type=str, default="")  # List - Directories
    parser.add_argument("-t", "--trajectories", nargs="+", type=str, default="")
    parser.add_argument("-fr", "--force_reload", action="store_true")
    parser.add_argument("-sd", "--save_data", action="store_false")

    parser.add_argument("-ss", "--sample_size", default=10, type=int)
    parser.add_argument("-indices", "--indices", default=None)
    parser.add_argument("-template", "--template", default=None)
    parser.add_argument("-cbf", "--covalent_bond_scale_factor", type=float, default=1.3)
    parser.add_argument("-angle", "--angle_from_central_atom", type=float, default=20.0)
    parser.add_argument("-n_points", "--number_of_points", type=int, default=8)

    args = parser.parse_args()
    directories = args.directories
    trajectories = args.trajectories

    save_data = args.save_data
    force_reload = args.force_reload

    sample_size = args.sample_size

    covalent_bond_scale_factor = args.covalent_bond_scale_factor
    n_points = args.number_of_points
    angle = args.angle_from_central_atom
    template = args.template
    indices = args.indices

    descriptor = Voronoi(
        indices=indices,
        template=template,
        covalent_bond_scale_factor=covalent_bond_scale_factor,
        n_points=n_points,
        angle_from_central_atom=angle,
        environment=None,
    )

    A = Analysis(descriptor, directories, sample_size, save_data=save_data, force_reload=force_reload)
    if len(directories) != 0:
        A.analyse_directories()

    if len(trajectories) != 0:
        print("Loading structures from trajectories")
        structures = []
        for trajectory in trajectories:
            structures += read(trajectory, index=":")

        A.structures += structures

    if len(directories) == 0 and len(trajectories) == 0:
        print("No directories or trajectories have been provided.")
        return

    print("\nConstructing combined sample from all directories and trajectories.\n")
    A.sample_size = 10000
    combined_unique = A.sort_structures(A.structures)
    write("combined_sample.traj", combined_unique)


if __name__ == "__main__":
    analysis()



import numpy as np

from agox.acquisitors.ABC_acquisitor import AcquisitorBaseClass
from agox.candidates import Candidate


class ReplicaExchangeAcquisitor(AcquisitorBaseClass):
    name = "ReplicaExchangeAcquisitor"

    def __init__(
        self,
        filter_last: bool = True,
        random: bool = True,
        sample_size: int = 10,
        gets: dict[str, str] = {"get_key": "sample"},
        always_lowest: bool = False,
        **kwargs,
    ) -> None:
        """
        Acquisitor for mananging selection of candidates for DFT evaluation in a
        replica exchange simulation.

        Parameters
        ----------
        dft_interval : int
            Number of iterations between DFT evaluations.
        filter_last : bool, optional
            If True, the last candidate in the list will be set to 0.0 fitness,
            which means it will not be selected for DFT evaluation. 
        random : bool, optional
            If True, the fitness values will be randomly generated between -1.0 and 0, otherwise 
            the fitness will be proportional to the index of walker it comes from, favouring low temperature walkers.
        sample_size : int, optional
            Number of candidates to sample from the pool of candidates. Default is 10.                
        """


        super().__init__(gets=gets, **kwargs)
        self.random = random
        self.always_lowest = always_lowest
        self.filter_last = filter_last
        self.sample_size = 10

    def calculate_acquisition_function(self, candidates: list[Candidate]) -> np.ndarray:
        if self.random:
            fitness = np.random.uniform(low=-1.0, high=0.0, size=len(candidates))
        else:
            fitness = np.linspace(-1.0, 0.0, num=len(candidates))
        should_ignore = False

        if len(candidates) >= self.sample_size:
            should_ignore = True
        if self.always_lowest:
            fitness[0] = -1.0
        if self.filter_last * should_ignore:
            fitness[len(candidates) - 1] = 0.0

        return fitness

    @classmethod
    def with_interval(cls, dft_interval: int, **kwargs):
        """
        Returns the interval for DFT evaluations.

        Parameters
        ----------
        dft_interval : int
            Number of iterations between DFT evaluations.

        Returns
        -------
        dict[str, int]
            Dictionary with the key 'dft_interval' and the value of the input parameter.
        """

        def skip_function(self, candidate_list: list[Candidate]) -> bool:
            """
            Function defining how often iterations are skipped for evaluation
            """
            if (self.get_iteration_counter()) % dft_interval != 0:
                return True
            else:
                return False

        return cls(
            skip_function=skip_function,
            **kwargs,
        )



import traceback
from dataclasses import dataclass, field
from typing import Callable

import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator
from ase.optimize.optimize import Optimizer

from agox.candidates import Candidate
from agox.candidates.ABC_candidate import disable_cache
from agox.models import Model


@dataclass
class RelaxationResult:
    candidate: Candidate
    relaxation_steps: int
    max_force: float
    initial_energy: float
    final_energy: float
    converged: bool
    message: list[str] = field(default=None)
    trajectory: list[Candidate] = field(default=None)


def remote_relax(
    model: Model,
    candidate: Candidate,
    optimizer: Optimizer,
    optimizer_kwargs: dict,
    optimizer_run_kwargs: dict,
    message_function: Callable | None = None,
    trajectory: bool = False,
) -> RelaxationResult:
    """
    Relaxes a candidate using the given model and optimizer. Purpose of this function 
    is to be used as a remote function in Ray.

    Parameters
    ----------
    model : agox.models.ABC_model
        Model to use for relaxation.
    candidate : agox.candidates.ABC_candidate
        Candidate to relax.
    optimizer : ase.optimize.Optimizer
        Optimizer to use for relaxation.
    optimizer_kwargs : dict
        Keyword arguments to pass to the optimizer constructor.
    optimizer_run_kwargs : dict
        Keyword arguments to pass to the optimizer.run method.
    message_function : callable
        Function that creates a message that will be printed. The function
        must takes: candidate, optimizer, model as arguments and return a string or
        a list of strings.
    trajectory: bool
        Whether or not trajectories are returned by the remote relaxation function.    
    """
    # Setup the candidate
    _candidate = candidate.copy()
    _candidate.set_calculator(model)
    
    disable_cache(_candidate)

    # Setup the ASE optimizer
    optimizer = optimizer(_candidate, **optimizer_kwargs)

    if trajectory: # Add callback to store trajectory
        _trajectory = []

        def trajectory_callback(atoms: Candidate) -> None:
            traj_atoms = atoms.copy()
            traj_atoms.set_constraint([]) # Make sure to get rid of BoxConstraint.
            calculator = SinglePointCalculator(traj_atoms, **atoms.calc.results)
            traj_atoms.calc = calculator
            _trajectory.append(traj_atoms)

        optimizer.attach(trajectory_callback, interval=1, atoms=candidate)
    else:
        _trajectory = None

    # Run relaxation
    try:
        initial_energy = _candidate.get_potential_energy()
        optimizer.run(**optimizer_run_kwargs)
        n_steps = optimizer.get_number_of_steps()
        gradient = -1 * _candidate.get_forces().flatten()
        converged = optimizer.converged(gradient=gradient)
        final_energy = _candidate.get_potential_energy()
    except Exception as e:
        traceback.print_exc()
        n_steps = -1
        converged = False
        initial_energy = None
        final_energy = None

    _candidate.add_meta_information("relaxation_steps", n_steps)

    # Calculate max force
    forces = _candidate.get_forces()
    max_force = np.linalg.norm(forces, axis=-1).max()

    if message_function is not None: # Create message if message function is given
        message = message_function(_candidate, optimizer, model)
        if not isinstance(message, list):
            message = [message]
    else:
        message = None

    result = RelaxationResult(
        candidate=_candidate,
        relaxation_steps=n_steps,
        max_force=max_force,
        converged=converged,
        message=message,
        trajectory=_trajectory,
        initial_energy=initial_energy,
        final_energy=final_energy,
    )
    return result



global mpl_backend
mpl_backend = "Agg"


def get_mpl_backend():
    global mpl_backend
    return mpl_backend


def set_mpl_backend(new_backend):
    global mpl_backend
    mpl_backend = new_backend


def use_agox_mpl_backend():
    import matplotlib

    matplotlib.use(get_mpl_backend())



from copy import deepcopy
from typing import List, Literal, Optional, Sequence, TypeAlias

try: 
    from typing import Self
except ImportError:
    Self: TypeAlias = "ConstraintManager"

import numpy as np
from ase.constraints import IndexedConstraint

from agox.candidates import CandidateBaseClass

Candidate: TypeAlias = CandidateBaseClass


class Constraint:
    """
    Type alias for ASE constraints.
    """


class ManagedConstraint:
    def __init__(
        self,
        constraint: Constraint,
        mode: Optional[Literal["all", "optimize", "template"]] = None,
        index: Optional[Sequence[int]] = None,
    ) -> None:
        """
        A wrapper around a constraint that manages the indices that the constraint is applied to.
        This helps for searches where candidates with a different number of atoms may be generated,
        as ASE's implementation of constraints don't take this into account.

        Parameters
        ----------

        constraint : Constraint
            The constraint to manage.
        index : Optional[Sequence[int]]
            The indices to apply the constraint to.
        mode : Optional[Literal["all", "optimize", "template"]]
            The mode in which to apply the constraint. If mode is not None, index is ignored.
            If mode is 'all' then the constraint is applied to all atoms.
            If mode is 'optimize' then the constraint is applied to the atoms that are being optimized.
            If mode is 'template' then the constraint is applied to the atoms that are in the template.

        """
        self.constraint = constraint
        self.index = index
        if mode in ["all", "optimize", "template"]:
            self.mode = mode
        else:
            self.mode = None
        self.is_index_constraint = isinstance(constraint, IndexedConstraint)

    def get(self, candidate: Candidate) -> Constraint:
        constraint = deepcopy(self.constraint)
        if self.is_index_constraint:
            index = self.get_index(candidate)
            constraint.index = index
        return constraint

    def get_index(self, candidate: Candidate) -> Sequence[int]:
        if self.mode is not None:
            index = self.get_index_from_mode(candidate)
        elif self.index is not None:
            index = self.index.copy()
        return index

    def get_index_from_mode(self, candidate: Candidate) -> Sequence[int]:
        if self.mode == "all":
            return np.arange(len(candidate))
        elif self.mode == "optimize":
            indices = candidate.get_optimize_indices()
            return indices
        elif self.mode == "template":
            indices = candidate.get_template_indices()
            return indices

    def __str__(self) -> str:
        string = f"ManagedConstraint({self.constraint.__class__.__name__})"
        if self.is_index_constraint:
            if self.index is not None:
                string += f" applied to {self.index}"
            elif self.mode is not None:
                string += f" applied to {self.mode}"
        return string


class ConstraintManager:
    def __init__(self) -> None:
        """
        A class to manage constraints for candidates.

        Effecitvely acts like a list of constraints, but with the ability to
        indicate the mode in which the constraint is applied.
        """
        self.constraints = []

    def add_constraint(
        self, constraint: Constraint, mode: Optional[Literal["all", "optimize", "template"]] = None
    ) -> None:
        if isinstance(constraint, IndexedConstraint):
            if mode is None:
                index = constraint.index.copy()
            else:
                index = None
            managed_constraint = ManagedConstraint(constraint, index=index, mode=mode)
        else:
            managed_constraint = ManagedConstraint(constraint)

        self.constraints.append(managed_constraint)

    def get_constraints(self, candidate: Candidate) -> List[Constraint]:
        constraints = []
        for managed_constraint in self.constraints:
            constraint = managed_constraint.get(candidate)
            constraints.append(constraint)
        return constraints

    def apply(self, candidate: Candidate) -> Candidate:
        constraints = self.get_constraints(candidate)
        candidate.set_constraint(constraints)
        return candidate

    def append(self, constraint: Candidate, mode: Optional[Literal["all", "optimize", "template"]] = None) -> None:
        self.add_constraint(constraint, mode)

    def __iter__(self) -> Self:
        self.iter_count = 0
        return self

    def __add__(self, other: Self) -> Self:
        new_manager = deepcopy(self)
        if isinstance(other, list):
            for constraint in other:
                new_manager.add_constraint(constraint)
        elif isinstance(other, ConstraintManager):
            for key, constraint in other.constraints.items():
                new_manager.add_constraint(constraint.constraint, constraint.index)

        return new_manager

    def _radd__(self, other: Self) -> Self:
        return self.__add__(other)

    def __next__(self) -> ManagedConstraint:
        if self.iter_count < len(self.constraints):
            constraint = self.constraints[self.iter_count]
            self.iter_count += 1
            return constraint
        else:
            raise StopIteration

    def __repr__(self) -> str:
        rep = """ConstraintManager:"""
        for constraint in self.constraints:
            rep += f"\n\t{constraint}"
        return rep



from typing import Optional

import numpy as np
from ase import Atoms
from ase.calculators.calculator import Calculator

from agox.models import CompositionModel, ModelBaseClass
from agox.utils import candidate_list_comprehension


class CalculatorModel(ModelBaseClass):
    """
    A model that turns an ASE calculator into a model that can be used with AGOX.
    """

    name = "CalculatorModel"
    implemented_properties = ["energy", "forces"]

    def __init__(self, calculator: Calculator, composition_model: bool = True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.calculator = calculator
        if composition_model:
            self.composition_model = CompositionModel()
        else:
            self.composition_model = None
            
        self.name = f"{calculator.__class__.__name__}Model"

    @candidate_list_comprehension
    def predict_energy(self, atoms: Atoms, **kwargs) -> float:
        energy = self.calculator.get_potential_energy(atoms)
        if self.composition_model is not None:
            energy += self.composition_model.predict_energy(atoms)
        return energy

    @candidate_list_comprehension
    def predict_forces(self, atoms: Atoms, **kwargs) -> np.ndarray:
        return self.calculator.get_forces(atoms)

    def train(
        self, data: list[Atoms], self_labels: Optional[np.ndarray] = None, truth_labels: Optional[np.ndarray] = None
    ) -> None:
        if self_labels is None:
            self.composition_model.reset()
            self_labels = np.array([self.predict_energy(atoms) for atoms in data])
        if truth_labels is None:
            truth_labels = np.array([atoms.get_potential_energy() for atoms in data])

        self.writer(f'Training composition model on {len(data)} configurations.')
        self.composition_model.train(data, energies = truth_labels - self_labels)

        indices = np.argwhere(self.composition_model.atom_energies != 0).flatten()
        self.writer("Composition model atom energies:")
        for i in indices:
            self.writer(f"Z={i}: {self.composition_model.atom_energies[i]:0.3f}")






import numpy as np

from agox.analysis import SearchData
from agox.analysis.property import ArrayPropertyData, FreeEnergyProperty


def test_energy_property_computes(free_energy_array_prop: ArrayPropertyData) -> None:
    print(free_energy_array_prop)

def test_energy_property_hasdata(free_energy_array_prop: ArrayPropertyData) -> None:
    assert free_energy_array_prop.data is not None

def test_energy_property_data_nparray(free_energy_array_prop: ArrayPropertyData) -> None:
    assert isinstance(free_energy_array_prop.data, np.ndarray)

def test_energy_property_shape(free_energy_array_prop: ArrayPropertyData) -> None:
    assert free_energy_array_prop.shape == ('Restarts', 'Indices [#]') or free_energy_array_prop.shape == ('Restarts', 'Iterations [#]')

def test_energy_property_axis(free_energy_array_prop: ArrayPropertyData) -> None:
    assert isinstance(free_energy_array_prop.axis, tuple)
    assert isinstance(free_energy_array_prop.axis[0], list)
    assert isinstance(free_energy_array_prop.axis[1], np.ndarray)

def test_energy_property_name(free_energy_array_prop: ArrayPropertyData) -> None:
    assert hasattr(free_energy_array_prop, 'name')

def test_energy_property_minimum(free_energy_property: FreeEnergyProperty, free_energy_array_prop: ArrayPropertyData, search_data: SearchData) -> None:
    min_energy = free_energy_property.get_minimum([search_data])
    np.testing.assert_allclose(free_energy_array_prop.data.min(), min_energy)




import numpy as np

from agox.utils.thermodynamics import ThermodynamicsData, gibbs_free_energy

from .property import ArrayPropertyData, Property, SearchData


class FreeEnergyProperty(Property):
    def __init__(self, thermo_data: ThermodynamicsData, time_axis=None):
        """
        Get the energy of the system as a np.array of shape [restarts, iterations].
        """
        if time_axis is None:
            time_axis = "indices"

        if time_axis not in ["indices", "iterations"]:
            raise ValueError("time_axis must be one of ['indices', 'iterations']")
        self.thermo_data = thermo_data
        self.time_axis = time_axis

    def get_time_axis(self, search_data: SearchData) -> str:
        axis_name = self.time_axis.capitalize()

        if self.time_axis == "indices":
            indices = search_data.get_all("indices", fill=np.nan)
        elif self.time_axis == "iterations":
            indices = search_data.get_all("iterations", fill=np.nan)
        else:
            indices = search_data.get_all("times", fill=np.nan) / self.unit * self.cost_factor
            if self.cost_factor == 1:
                axis_name = axis_name + f" [{self.time_unit_symbol}]"
            else:
                axis_name = f"CPU cost [{self.time_unit_symbol}]"

        if self.time_axis in ["indices", "iterations"]:
            axis_name = f"{axis_name} [#]"

        return axis_name, indices

    def compute(self, search_data) -> np.array:
        """
        Get the energy of the system as a np.array of shape [restarts, iterations].
        """

        axis_name, indices = self.get_time_axis(search_data)

        candidates = search_data.get_all_candidates()
        energy = search_data.get_all("energies", fill=np.inf)
        free_energy = np.ones_like(energy) * np.inf

        N_restarts = search_data.get_number_of_restarts()

        for restart in range(N_restarts):
            for observation in range(len(candidates[restart])):
                candidate = candidates[restart][observation]
                free_energy[restart, observation] = gibbs_free_energy(candidate=candidate, thermo_data=self.thermo_data)

        free_energy_property = ArrayPropertyData(
            data=free_energy,
            name='Free Energy',
            shape=("Restarts", axis_name),
            array_axis=(search_data.get_all_identifiers(), indices),
        )

        return free_energy_property



from agox.samplers.replica_exchange.rate_tracker import RateTracker
from agox.samplers.replica_exchange.sample import ReplicaExchangeSample
from agox.samplers.replica_exchange.replica_exchange import ReplicaExchangeSampler

__all__ = ['RateTracker', 'ReplicaExchangeSample', 'ReplicaExchangeSampler']


from timeit import default_timer as dt
from typing import Any, Callable, Dict, List, Optional, Union
from uuid import uuid4


class ObserverMethod:
    def __init__(
        self,
        class_name: str,
        method_name: str,
        method: Callable,
        gets: Dict,
        sets: Dict,
        order: Union[float, int],
        handler_identifier: str,
    ) -> None:
        """
        ObserverMethod class. Holds all information about the method that an
        observer class attaches to the observer-loop.

        Parameters
        ----------
        name: str
            Name of the class
        method_name: str
            Name of the method.
        method: method
            The method that is called by the observer-loop.
        sets: dict
            Dict containing the keys that the method will set in the cache.
        gets: dict
            Dict containing the keys that the method will get from the cache.
        order: float
            Execution order
        """
        self.class_name = class_name
        self.method_name = method_name
        self.method = method
        self.gets = gets
        self.sets = sets
        self.order = order

        self.name = self.class_name + "." + self.method_name
        self.class_reference = method.__self__
        self.key = uuid4()
        self.handler_identifier = handler_identifier
        self.handler_name = None

    def __getitem__(self, key: str) -> Union[str, Callable, Dict]:
        return self.__dict__[key]

    def __call__(self, *args, **kwargs) -> Any:
        state = kwargs.get("state")
        state.execution_times[self.get_name()] = -dt()
        result = self.method(*args, **kwargs)
        state.execution_times[self.get_name()] += dt()
        return result

    def get_name(self, handler_name: bool = True, class_name: bool = True, method_name: bool = True) -> str:
        name = ""
        if handler_name:
            name += self.handler_name + "."
        if class_name:
            name += self.class_name + "."
        if method_name:
            name += self.method_name
        return name

    def set_handler(self, handler_name: str) -> None:
        self.handler_name = handler_name

    def report(
        self,
        offset: Optional[str] = None,
        report_key: bool = False,
        return_report: bool = False,
        print_report: bool = True,
    ) -> Union[List[str], None]:
        """
        Generate report. Used by ObserverHandler.observer_reports

        Parameters
        ----------
        offset: str
        report_key: bool
            Print the key or not.
        return_report: bool
            Whether to return the report or not.
        print_report: bool
            Whether to print hte print or not.

        Returns
        -------
        List[str] or None:
            If report_key = True then it returns a string otherwise None is returned.
        """
        report = []
        if offset is None:
            offset = ""

        for key in self.gets.keys():
            value = self.gets[key]
            out = offset + f"Gets '{value}'"
            if report_key:
                out += f" using key attribute '{key}'"
            report.append(out)

        for key in self.sets.keys():
            value = self.sets[key]
            out = offset + f"Sets '{value}'"
            if report_key:
                out += f" using key attribute '{key}'"
            report.append(out)

        if len(self.gets) == 0 and len(self.sets) == 0:
            out = offset + "Doesnt set/get anything"
            report.append(out)

        if print_report:
            for string in report:
                print(string)
        if return_report:
            return report



from pathlib import Path

import numpy as np
from ase import Atoms
from ase.io import read
from ase.optimize import FIRE

from agox import AGOX
from agox.acquisitors import GibbsFreeEnergyAquisitor
from agox.collectors import ParallelCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import AddRemoveGenerator, RandomGenerator, RattleGenerator
from agox.helpers import SubprocessGPAW
from agox.models.descriptors.soap import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.postprocessors import ParallelRelaxPostprocess
from agox.postprocessors.minimum_dist import MinimumDistPostProcess
from agox.samplers import GFEKMeansSampler
from agox.utils.sparsifiers import CUR
from agox.utils.thermodynamics import ThermodynamicsData

##############################################################################
# General settings:
##############################################################################

# Manually set seed and database-index - for production searches with repeated independent runs the seed should be be set to different values.
seed = 41
database_index = 0

# Candidate gnerator settings:
# Number of candidates to generate for each iteration with each generator: random, rattle, add/remove.
# For efficiency the total number should be an integer multiple of the number of the number of CPUs used.
generator_dictionary = {0: [32, 0, 0], 5: [6, 26, 0], 10: [4, 20, 8]}
C_dict = {"c1": 0.85, "c2": 1.5}  # Constants for the generators, used in rattle and add/remove generators.

# Model relaxations settings:
model_relax_steps = 150  # Number of steps to relax the surrogate model.
model_fmax = 0.05  # Force convergence criterion for the model relaxations in eV/Angstrom.

N_iterations = 500  # Number of iterations to run the AGOX loop.

##############################################################################
# Grand Canonical Global Optimization settings:
##############################################################################
"""
For a DFT search it is convenient to not calculate the references in the same script, but rather load them from a file.
The `ThermodynamicsData` object can be instantiated from a json-file, that looks like this:
{
    "references": {
        "O": -2.970580729899092,
        "Pd": -3.936261511645199
    },
    "chemical_potentials": {
        "O": 0.0,
        "Pd": 0.0
    },
    "template_energy": -30.56738172286689
}
"""
thermo_data_path = ...
symbols_range = {"Pd": [4, 5], "O": [0, 1, 2, 3, 4]}  # Range of stoichiometries for the elements in the system.
mu_O = -1.0  # Chemical potential for oxygen in eV can be set to the value desired for the search.

##############################################################################
# System & general settings:
##############################################################################

data_directory = Path("/home/machri/Projects/agox/runscripts/gc_jon/original/scripts/palladium_oxide/v8/template")
template = read(data_directory / "Pd100_sqrt5_n2.traj")

confinement_cell = template.get_cell()
confinement_cell[2, 2] = 4
z0 = template.positions[:, 2].max()
confinement_corner = np.array([0, 0, z0])

environment = Environment(
    template=template,
    symbols_range=symbols_range,
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[True, True, False],
)

# Database
db_path = "db{}.db".format(database_index)
database = Database(filename=db_path, order=7)

##############################################################################
# Calculator
##############################################################################

settings = {"mode": {"name": "pw", "ecut": 400}, "xc": "PBE", "kpts": (2, 2, 1), "txt": "gpaw.log"}

calc = SubprocessGPAW(**settings)

################################################################################
# Reference energies.
################################################################################

thermo_data = ThermodynamicsData.from_file(thermo_data_path)
thermo_data.set_chemical_potential("O", mu_O)
thermo_data.save("gfe.json")

##############################################################################
# Search Settings:
##############################################################################

descriptor = SOAP(
    environment=environment,
    r_cut=6,
    nmax=3,
    lmax=2,
    sigma=1,
    weight=True,
    periodic=True,
)
kernel = C(5, (1, 100)) * RBF(20, (10, 30))

single_atom_energies = np.zeros(100)
for symbol in thermo_data.references.keys():
    number = Atoms(symbol).numbers[0]
    single_atom_energies[number] = thermo_data.references[symbol]

model = SparseGPR(
    descriptor=descriptor,
    kernel=kernel,
    database=database,
    sparsifier=CUR(1000),  # sparsification using CUR algorithm with 1000 sparse points
    sigma=0.01,  # eV/atom
    prior=Repulsive(ratio=0.8),
    use_ray=False,
    train_uncertainty=True,
    single_atom_energies=single_atom_energies,
)

sample_size = 10
sampler = GFEKMeansSampler(
    thermo_data=thermo_data,
    model=model,
    descriptor=descriptor,
    database=database,
    sample_size=sample_size,
)
rattle_generator = RattleGenerator(**C_dict, **environment.get_confinement())
random_generator = RandomGenerator(**C_dict, **environment.get_confinement(), contiguous=False)
addrm_generator = AddRemoveGenerator(**C_dict, **environment.get_confinement())

# Dict specificies how many candidates are created with and the dict-keys are iterations.
generators = [random_generator, rattle_generator, addrm_generator]
num_candidates = generator_dictionary

acquisitor = GibbsFreeEnergyAquisitor(
    thermo_data=thermo_data,
    model=model,
    kappa=2,
    order=3,
)
# CPU-count is set here for Ray - leave it out to use as many cores as are available.
collector = ParallelCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)
# Number of steps is very low - should be set higher for a real search!
relaxer = ParallelRelaxPostprocess(
    model=acquisitor.get_acquisition_calculator(),
    constraints=environment.get_constraints(),
    optimizer=FIRE,
    optimizer_run_kwargs={"steps": model_relax_steps, "fmax": model_fmax},
    start_relax=10,
    order=2,
)

min_dist = MinimumDistPostProcess(order=2.5, **C_dict)

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "prioritized_candidates"},
    optimizer_kwargs={"logfile": None},
    store_trajectory=True,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 0},
    order=6,
)
##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(collector, acquisitor, relaxer, database, evaluator, min_dist, seed=seed)

agox.run(N_iterations=N_iterations)



import os

import rich_click as click
from ase.io import read, write
from rich import print

from agox.databases.database import Database


class Analysis:
    def __init__(self, descriptor, directories, sample_size, save_data=True, force_reload=False):
        self.descriptor = descriptor
        self.directories = directories
        self.sample_size = sample_size
        self.analysed_directories = []

        self.save_data = save_data
        self.force_reload = force_reload

        self.structures = []

    def analyse_directories(self):
        for idx, directory in enumerate(self.directories):
            self.add_directory(directory, force_reload=self.force_reload)

    def add_directory(self, directory, force_reload=False):
        if directory in self.analysed_directories:
            print("Directory: {} - already been loaded so skipping it".format(directory))

        cwd = os.getcwd()
        full_path = os.path.join(cwd, directory)
        if full_path[-1] != "/":
            full_path += "/"
        data_path = full_path + "GS_data/"

        if os.path.exists(data_path) and not force_reload:
            print("{} has already been analysed.".format(full_path))
            self.structures += read(data_path + "unique.traj", index=":")
            if self.descriptor.template is None:
                template = read(data_path + "template.traj", index="0")
                self.descriptor.template = template

        else:
            # Find database files in this directory:
            structures = []
            print("Loading files from directory: {}".format(directory))
            for f in os.listdir(os.path.join(cwd, directory)):
                if f[-3:] == ".db":
                    db = Database(os.path.join(cwd, directory, f))
                    db.restore_to_memory()
                    structures += db.get_all_candidates()
                    if self.descriptor.template is None:
                        template = structures[0].get_template()
                        self.descriptor.template = template

            unique = self.sort_structures(structures)
            self.structures += unique

            if self.save_data:
                if not os.path.exists(data_path):
                    os.mkdir(data_path)
                write(data_path + "unique.traj", unique)
                write(data_path + "template.traj", self.descriptor.template)

        self.analysed_directories.append(directory)

    def sort_structures(self, structures):
        structures = sorted(structures, key=lambda x: x.get_potential_energy())
        print(f"Total number of structures: {len(structures)}")
        unique = []
        eigen_spectrum = []
        for i, structure in enumerate(structures):
            eigen_values = self.descriptor.create_features(structure)
            if eigen_values not in eigen_spectrum:
                unique.append(structure)
                eigen_spectrum.append(eigen_values)
                print(i, eigen_values)
                if len(unique) == self.sample_size:
                    break
        print(f"Number of unique structures: {len(unique)}")
        return unique


graph_sort_option_groups = {
    "agox graph-sort": [
        {
            "name": "File options",
            "options": ["--directories", "--trajectories", "--save-data", "--force-reload"],
        },
        {
            "name": "Sorting options",
            "options": ["--sample-size", "--covalent-bond-scale-factor"],
        },
        {
            "name": "Voronoi options",
            "options": [
                "--number-of-points",
                "--angle-from-central-atom",
                "--template",
                "--indices",
            ],
        },
    ]
}

click.rich_click.OPTION_GROUPS.update(graph_sort_option_groups)


@click.command("graph-sort")
@click.option(
    "--directories",
    "-d",
    multiple=True,
    help="Directories to analyse.",
    required=False,
    default=[],
)
@click.option(
    "--trajectories",
    "-t",
    multiple=True,
    help="Trajectories to analyse.",
    required=False,
    default=[],
)
@click.option(
    "--save-data",
    "-sd",
    is_flag=True,
    help="Save the data to a file.",
    required=False,
    default=True,
)
@click.option(
    "--force-reload",
    "-f",
    is_flag=True,
    help="Force reload the data.",
    required=False,
    default=False,
)
@click.option(
    "--sample-size",
    "-ss",
    type=int,
    help="Number of structures to sample.",
    required=False,
    default=10,
)
@click.option(
    "--covalent-bond-scale-factor",
    "-cbf",
    type=float,
    help="Covalent bond scale factor.",
    required=False,
    default=1.3,
)
@click.option(
    "--number-of-points",
    "-n",
    type=int,
    help="Number of points in the Voronoi diagram.",
    required=False,
    default=8,
)
@click.option(
    "--angle-from-central-atom",
    "-a",
    type=float,
    help="Angle from the central atom.",
    required=False,
    default=20.0,
)
@click.option("--template", "-tm", type=str, help="Template file.", required=False, default=None)
@click.option(
    "--indices",
    "-i",
    type=str,
    help="Indices to consider.",
    required=False,
    default=None,
)
def cli_graph_sorting(
    directories,
    trajectories,
    save_data,
    force_reload,
    sample_size,
    covalent_bond_scale_factor,
    number_of_points,
    angle_from_central_atom,
    template,
    indices,
):
    """
    Sorts structures according to their Voronoi graphs.
    """
    from agox.models.descriptors.voronoi import Voronoi

    descriptor = Voronoi(
        indices=indices,
        template=template,
        covalent_bond_scale_factor=covalent_bond_scale_factor,
        n_points=number_of_points,
        angle_from_central_atom=angle_from_central_atom,
        environment=None,
    )

    A = Analysis(
        descriptor,
        directories,
        sample_size,
        save_data=save_data,
        force_reload=force_reload,
    )
    if len(directories) != 0:
        A.analyse_directories()

    if len(trajectories) != 0:
        print("Loading structures from trajectories")
        structures = []
        for trajectory in trajectories:
            structures += read(trajectory, index=":")

        A.structures += structures

    if len(directories) == 0 and len(trajectories) == 0:
        print("No directories or trajectories have been provided.")
        return

    print("\nConstructing combined sample from all directories and trajectories.\n")
    A.sample_size = 10000
    combined_unique = A.sort_structures(A.structures)
    write("combined_sample.traj", combined_unique)


if __name__ == "__main__":
    cli_graph_sorting()




import numpy as np

from agox.models.descriptors.type_descriptor import TypeDescriptor
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import DotProduct
from agox.models.GPR.priors import Repulsive
from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass


class OnehotUniqueSparsifier(SparsifierBaseClass):
    name = "OnehotUnique"

    def __init__(self, n_species):
        self.m_points = n_species

    def sparsify(self, X: np.ndarray) -> np.ndarray:
        return np.eye(self.m_points), None


class LocalMean(SparseGPR):
    name = "LocalMeanModel"

    def __init__(self, species, database=None, prior=Repulsive(ratio=0.7), **kwargs):
        kernel = DotProduct(0)
        descriptor = TypeDescriptor.from_species(species)
        sparsifier = OnehotUniqueSparsifier(n_species=len(species))
        super().__init__(
            kernel=kernel,
            descriptor=descriptor,
            noise=0.001,
            use_ray=False,
            n_optimize=0,
            prior=prior,
            sparsifier=sparsifier,
            database=database,
        )

        self.dynamic_attributes = ["X", "Xm", "Kmm_inv", "alpha", "C_inv", "kernel", "mean_energy"]

    def predict_forces(self, atoms, **kwargs):
        F = np.zeros((len(atoms), 3))
        return self._postprocess_forces(atoms, F)

    def train(self, data):
        super().train(data)

        self.writer("Energies for each species:")
        for alpha, species in zip(self.alpha.flatten(), self.descriptor.unique_numbers):
            self.writer(f"    Atomic number {species}: {alpha:.3f}")



from agox.analysis.search_data import SearchCollection, SearchData  # noqa
from agox.analysis.property import (
    ArrayPropertyData,
    DescriptorProperty,
    EnergyProperty,
    FreeEnergyProperty,
    ListPropertyData,
)
from agox.analysis.criterion import BaseCriterion, DistanceCriterion, ThresholdCriterion

from agox.analysis.plot import PropertyPlotter, SuccessPlotter

__all__ = [
    "SearchCollection",
    "SearchData",
    "ArrayPropertyData",
    "DescriptorProperty",
    "EnergyProperty",
    "FreeEnergyProperty",
    "ListPropertyData",
    "BaseCriterion",
    "DistanceCriterion",
    "ThresholdCriterion",
    "PropertyPlotter",
    "SuccessPlotter",
]



import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, List, Optional
from uuid import uuid4

import numpy as np
from ase.data import covalent_radii
from ase.geometry import get_distances

from agox.candidates import Candidate, StandardCandidate
from agox.environments import Environment
from agox.helpers.confinement import Confinement
from agox.main import State
from agox.observer import Observer
from agox.samplers import Sampler

dimensionality_angles = {
    3: {"theta": [0, 2 * np.pi], "phi": [0, np.pi]},
    2: {"theta": [0, 2 * np.pi], "phi": [np.pi / 2, np.pi / 2]},
    1: {"theta": [0, 0], "phi": [np.pi / 2, np.pi / 2]},
}


class GeneratorBaseClass(ABC, Observer):
    def __init__(
        self,
        confinement: Optional[Confinement] = None,
        confinement_cell: Optional[np.ndarray] = None,
        confinement_corner: Optional[np.ndarray] = None,    
        c1: float = 0.75,
        c2: float = 1.25,
        use_mic: bool = True,
        environment: Sampler = None,
        sampler: Environment = None,
        gets: dict[str, str] | None = None,
        sets: dict[str, str] | None = None,
        order: int = 2,
        replace: bool = True,
        surname: Optional[str] = None,
        **kwargs,
    ) -> None:
        """ 
        Base class for all generators.

        Parameters:
        -----------
        confinement: Confinement object
            Confinement object that the generator is confined to.
        confinement_cell: np.ndarray
            The cell that the generator is confined to. Only used if confinement is None.
        confinement_corner: np.ndarray
            Lower left corner of the confinement cell. Only used if confinement is None.
        c1: float
            Lower limit on bond lengths.
        c2: float
            Upper limit on bond lengths.
        use_mic: bool
            If the minimum image convention should be used.
        environment: Environment object
            The environment object used in the search.
        sampler: Sampler object
            Sampler object used to create and get sample members from. Only 
            relevant if the generator is used without a collector.
        gets: Dict
            Dictionary of get methods.
        sets: Dict
            Dictionary of set methods.
        order: int
            Order of the generator.
        replace: bool
            If the sampler should replace the parents.
        """
        if sets is None:
            sets = {"set_key": "candidates"}
        Observer.__init__(self, sets=sets, gets=gets, order=order, surname=surname, **kwargs)

        self.c1 = c1  # Lower limit on bond lengths.
        self.c2 = c2  # Upper limit on bond lengths.
        self.use_mic = use_mic
        self.environment = environment
        self.sampler = sampler
        self.replace = replace
        self.candidate_instanstiator = StandardCandidate

        if confinement is not None:
            self.confinement = confinement
        else:
            self.confinement = Confinement(cell=confinement_cell, corner=confinement_corner)

        if self.environment is not None:
            self.plot_confinement(environment)

        self.add_observer_method(
            self.generate, sets=self.sets[0], gets=self.gets[0], order=self.order[0], handler_identifier="AGOX"
        )

    @abstractmethod
    def _get_candidates(
        self, candidate: Candidate, parents: List[Candidate], environment: Environment
    ) -> List:  # pragma: no cover
        """
        Method that returns a list of generated candidates
        """
        return []

    def get_candidates(self, sampler: Sampler, environment: Environment) -> List[Candidate]:
        """
        Method used to get new candidates.
        A generator method must have 'get_number_of_parents' method implemented.

        Parameters:
        -----------
        sampler: Sampler object
            Sampler object used to create and get sample members from.
        environment: Environment object
            The environment object used in the search.

        Returns: list
            List of candidates generated by the generator.
        """
        number_of_parents = self.get_number_of_parents(sampler)

        # If no parents are needed, the search template is given to generator.
        if number_of_parents == 0:
            parents = []
            candidate = environment.get_template().copy()
        # Check that the returned number of parents equal the required number.
        else:
            parents = sampler.get_parents(number_of_parents, replace=self.replace)
            if len(parents) != number_of_parents:
                return []
            candidate = deepcopy(parents[0])  # Use the first parent as a template for the candidate.
            candidate.info = dict() # Info should not be used but this is a safe guard against propagating info from parents if it has been set.
            candidate.reset_meta_information()
            candidate.set_constraint()

        # Generates new candidates and add relevant meta information.
        candidates = self._get_candidates(candidate, parents, environment)
        for candidate in candidates:
            candidate.add_meta_information("generator", self.name)
            candidate.add_meta_information("uuid", str(uuid4()))
            candidate.add_meta_information(
                "parent_search_iterations", [p.get_meta_information("search_iteration") for p in parents]
            )
            candidate.add_meta_information("parent_uuids", [p.get_meta_information("uuid") for p in parents])

            if parents:
                spawn_uuids = []
                for parent in parents:
                    try:
                        spawn_uuids += parent.get_meta_information("spawn_uuids")
                    except TypeError:
                        spawn_uuid = parent.get_meta_information("spawn_uuids")
                        raise ValueError(
                            f"Parent candidate does not have 'spawn_uuids' meta information correctly set. {spawn_uuid}"
                        )
            else:
                spawn_uuids = [candidate.get_meta_information("uuid")]

            candidate.add_meta_information("spawn_uuids", spawn_uuids)

        return candidates

    @property
    @abstractmethod
    def name(self) -> str:  # pragma: no cover
        pass

    @abstractmethod
    def get_number_of_parents(self, sampler: Sampler) -> int:  # pragma: no cover
        """
        Method that returns the number of parents needed to generate a candidate.

        Parameters:
        -----------
        sampler: Sampler object
            Sampler object used to create and get sample members from.

        Returns: Int
            Number of parents needed to generate a candidate.
        """
        pass

    def __call__(self, sampler: Sampler, environment: Environment) -> List[Candidate]:
        return self.get_candidates(sampler, environment)

    ####################################################################################################################
    # Convenience methods:
    ####################################################################################################################

    def check_new_position(
        self,
        candidate: Candidate,
        new_position: np.ndarray,
        number: int,
        skipped_indices: Optional[List] = None,
    ) -> bool:
        """ """

        if skipped_indices is None:
            skipped_indices = []

        state = True
        succesful = False
        for i in range(len(candidate)):
            if i in skipped_indices:
                continue

            covalent_dist_ij = covalent_radii[candidate[i].number] + covalent_radii[number]
            rmin = self.c1 * covalent_dist_ij
            rmax = self.c2 * covalent_dist_ij

            if self.use_mic:
                vector, distance = get_distances(
                    new_position, candidate.positions[i], cell=candidate.cell, pbc=candidate.pbc
                )
            else:
                distance = np.linalg.norm(candidate.positions[i] - new_position)
            if distance < rmin:  # If a position is too close we should just give up.
                return False
            elif not distance > rmax:  # If at least a single position is not too far we have a bond.
                succesful = True
        return succesful * state

    def get_sphere_vector(self, atomic_number_i: int, atomic_number_j: int) -> np.ndarray:
        """
        Get a random vector on the sphere of appropriate radii.

        Behavior changes based on self.dimensionality:
        3: Vector on sphere.
        2: Vector on circle (in xy)
        1: Vector on line (x)
        """
        dim = self.confinement.dimensionality        
        covalent_dist_ij = covalent_radii[atomic_number_i] + covalent_radii[atomic_number_j]
        rmin = self.c1 * covalent_dist_ij
        rmax = self.c2 * covalent_dist_ij
        r = np.random.uniform(rmin**dim, rmax**dim) ** (1 / dim)
        return self.get_displacement_vector(r)

    def get_displacement_vector(self, radius: float) -> np.ndarray:
        dim = self.confinement.dimensionality        
        theta = np.random.uniform(*dimensionality_angles[dim]["theta"])
        phi = np.random.uniform(*dimensionality_angles[dim]["phi"])
        displacement = radius * np.array([np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)])
        return displacement

    def get_box_vector(self) -> np.ndarray:
        return self.confinement._get_box_vector()  # From confinement-class.
    
    def check_confinement(self, position: np.ndarray) -> bool:
        return self.confinement.check_confinement(position)
    
    def get_dimensionality(self) -> int:
        return self.confinement.dimensionality

    ####################################################################################################################
    # Observer functions
    ####################################################################################################################

    def start_candidate(self) -> List[StandardCandidate]:
        """
        This method generates a candidate using the start generator, which allows other generators
        to kick-start the sampler.
        """
        from agox.generators import RandomGenerator

        return RandomGenerator(
            confinement=self.confinement.copy(),
            c1=self.c1,
            c2=self.c2,
            use_mic=self.use_mic,
        )(self.sampler, self.environment)

    @Observer.observer_method
    def generate(self, state: State) -> None:
        candidates = self.get_candidates(self.sampler, self.environment)
        if len(candidates) == 0 and self.sampler is not None and not self.sampler.initialized:
            candidates = self.start_candidate()
            self.writer("Fall-back to start generator, generated {} candidate ".format(len(candidates)))
        state.add_to_cache(self, self.set_key, candidates, mode="a")

    def plot_confinement(self, environment: Environment) -> None:
        if not self.confinement.confined:
            return

        import matplotlib.pyplot as plt

        from agox.utils.matplotlib_utils import use_agox_mpl_backend
        from agox.utils.plot import plot_atoms, plot_cell

        use_agox_mpl_backend()

        atoms = environment.get_template()
        atoms.set_constraint(environment.get_constraints())

        fig, axs = plt.subplots(1, 3, figsize=(15, 5))

        for ax, plane in zip(axs, ["xy", "xz", "yz"]):
            plot_atoms(ax, atoms, plane=plane, plot_constraint=True)
            plot_cell(ax, atoms.cell, plane=plane, collection_kwargs=dict(edgecolors="black", linestyles="dotted"))
            plot_cell(
                ax,
                self.confinement.cell,
                plane=plane,
                offset=self.confinement.corner,
                collection_kwargs=dict(edgecolors="red", linestyles="dashed"),
            )

        fig.savefig(f"confinement_plot_{self.name}.png")
        plt.close()



import traceback

import numpy as np
from ase.calculators.calculator import all_properties
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms
from ase.optimize import BFGS

from agox.candidates.ABC_candidate import disable_cache
from agox.postprocessors.ABC_postprocess import PostprocessBaseClass
from agox.utils.ray import RayPoolUser, Task


def relax(model, candidate, optimizer, optimizer_kwargs, optimizer_run_kwargs):
    candidate = candidate.copy()
    candidate.set_calculator(model)
    disable_cache(candidate)
    optimizer = optimizer(candidate, **optimizer_kwargs)
    try:
        optimizer.run(**optimizer_run_kwargs)
    except Exception as e:
        traceback.print_exc()
        print("Relaxation failed with exception: {}".format(e))
        candidate.add_meta_information("relaxation_steps", -1)
        return candidate
    candidate.add_meta_information("relaxation_steps", optimizer.get_number_of_steps())
    return candidate


class ParallelRelaxPostprocess(PostprocessBaseClass, RayPoolUser):
    """
    Relaxes candidates in parallel using Ray.

    Parameters
    ----------
    model : agox.models.ABC_model
        Model to use for relaxation.
    optimizer : ase.optimize.Optimizer
        Optimizer to use for relaxation.
    optimizer_run_kwargs : dict
        Keyword arguments to pass to the optimizer.run method.
    optimizer_kwargs : dict
        Keyword arguments to pass to the optimizer constructor.
    fix_template : bool
        If True, the template atoms are fixed during relaxation.
    constraints : list
        List of constraints to apply during relaxation.
    start_relax : int
        The iteration to start relaxing candidates.
    """

    name = "PoolRelaxer"

    def __init__(
        self,
        model=None,
        optimizer=None,
        optimizer_kwargs={"logfile": None},
        optimizer_run_kwargs={"fmax": 0.05, "steps": 200},
        constraints=None,
        start_relax=1,
        **kwargs,
    ):
        RayPoolUser.__init__(self)
        PostprocessBaseClass.__init__(self, **kwargs)

        self.optimizer = BFGS if optimizer is None else optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.optimizer_run_kwargs = optimizer_run_kwargs
        self.start_relax = start_relax
        self.model = model
        self.constraints = constraints if constraints is not None else []
        self.model_key = self.pool_add_module(model)

    def process_list(self, candidates):
        """
        Relaxes the given candidates in parallel using Ray.

        Parameters
        ----------
        list_of_candidates : listn
            List of AGOX candidates to relax.

        Returns
        -------
        list
            List of relaxed candidates.
        """

        if len(candidates) == 0:
            return candidates

        # Apply constraints.
        [self.apply_constraints(candidate) for candidate in candidates]

        # Make args, kwargs and modules lists:
        tasks = []
        for candidate in candidates:
            # Create task: function, modules, args, kwargs
            task = Task(
                function=relax,
                modules=[self.model_key],  # The model is stored on the actors, tell the task where to find it.
                args=[candidate, self.optimizer, self.optimizer_kwargs, self.optimizer_run_kwargs],
                kwargs={},
            )
            tasks.append(task)

        relaxed_candidates = self.task_map(tasks)

        # Remove constraints & move relaxed positions to input candidates:
        # This is due to immutability of the candidates coming from pool_map.
        [self.remove_constraints(candidate) for candidate in candidates]
        for cand, relax_cand in zip(candidates, relaxed_candidates):
            cand.set_positions(relax_cand.positions)
            results = {prop: val for prop, val in relax_cand.calc.results.items() if prop in all_properties}
            cand.calc = SinglePointCalculator(cand, **results)
            cand.meta_information.update(relax_cand.info)

        # Remove if relaxation failed:
        candidates = [
            cand
            for cand, rcand in zip(candidates, relaxed_candidates)
            if rcand.get_meta_information("relaxation_steps") > 0
        ]

        steps = np.array([candidate.get_meta_information("relaxation_steps") for candidate in relaxed_candidates])
        self.writer(f"{len(relaxed_candidates)} candidates relaxed.")
        self.writer("   Average number of steps: {:.1f}".format(steps.mean()))
        self.writer("   Std. number of steps: {:.1f}".format(steps.std()))

        max_forces = np.array([np.linalg.norm(candidate.get_forces(), axis=1).max(axis=0) for candidate in relaxed_candidates]).mean()
        self.writer(f"   Mean max force: {max_forces.mean():.3f}")
        self.writer(f"   Std. max force: {max_forces.std():.3f}")

        return candidates

    def postprocess(self, candidate):
        raise NotImplementedError('"postprocess"-method is not implemented, use postprocess_list.')

    def do_check(self, **kwargs):
        return self.check_iteration_counter(self.start_relax)

    ####################################################################################################################
    # Constraints
    ####################################################################################################################

    def apply_constraints(self, candidate):
        candidate.set_constraint(self.constraints)

    def remove_constraints(self, candidate):
        candidate.set_constraint([])





from agox.evaluators.single_point import SinglePointEvaluator


class RattleEvaluator(SinglePointEvaluator):
    def __init__(self, calculator, seed=0, constraints=[], stdev=0.1, skip_frequency=10, **kwargs):
        super().__init__(calculator=calculator, **kwargs)
        self.seed = seed
        self.constraints = constraints
        self.stdev = stdev
        self.skip_frequency = skip_frequency

    def evaluate_candidate(self, candidate):
        if self.get_iteration_counter() % self.skip_frequency > 0:
            candidate.set_constraint(self.constraints)

            self.writer("Im rattling with stdev:", self.stdev)

            seed = self.seed * 10000 + self.get_iteration_counter()
            candidate.rattle(stdev=self.stdev, seed=seed)
            candidate.set_constraint()

        return super().evaluate_candidate(candidate)



import os

try:  # When installing the package we don't need to actually import.
    from agox.module import Module  # noqa
    from agox.observer import Observer
    from agox.writer import Writer
    from agox.cli.main import main
    from agox.main.state import State
    from agox.main.agox import AGOX

    __all__ = ["Module", "Observer", "Writer", "State", "AGOX", "__version__", "main"]

except ImportError as e:
    print(e)
    pass



import ray
import time
from timeit import default_timer as timer

ray.init(num_cpus=4)


@ray.remote
def f(x):
    time.sleep(1)
    return x * x


t0 = timer()
futures = [f.remote(i) for i in range(4)]
t1 = timer()
results = ray.get(futures)
t2 = timer()
print("Time elapsed t1-t0: ", t1 - t0)
print("Time elapsed t2-t1: ", t2 - t1)



from functools import reduce

import numpy as np
from ase import Atoms
from ase.atoms import symbols2numbers
from ase.symbols import Symbols

from agox.environments.ABC_environment import EnvironmentBaseClass


# This is a bit messy but it crates a matrix with a combination of all possible atom types and numbers so that it works without changing
# the underlying self._numbers approach
def make_numbers_list_from_matrix(symbols, matrix):
    numbers_list = []
    for line in matrix:
        combination = []
        for j in range(len(symbols)):
            combination += line[j] * [symbols2numbers(symbols[j])]
        combination = list(reduce(lambda x, y: x + y, combination, []))
        numbers_list.append(combination)
    return numbers_list


def set_combination_matrix(symbols_dict):
    symbols = list(symbols_dict.keys())
    lengths = 1
    lengths_array = []
    for i in symbols:
        lengths_array.append(len(list(symbols_dict[i])))
        lengths = lengths * len(list(symbols_dict[i]))
    combiantion_matrix = np.zeros((lengths, len(symbols)), dtype=int)
    for x in range(lengths):
        for i in range(len(symbols)):
            advancement = int(np.prod(lengths_array[i + 1 : len(symbols)]))
            pos = int(x / advancement) % len(symbols_dict[symbols[i]])
            combiantion_matrix[x, i] = symbols_dict[symbols[i]][pos]
    return make_numbers_list_from_matrix(symbols, combiantion_matrix)


class Environment(EnvironmentBaseClass):
    """
    Environment class for the generation of candidates.

    Parameters
    ----------
    template : Atoms
        The template structure to use.
    numbers : list, optional
        List of atomic numbers to add to the template, by default None.
    symbols : list, optional
        List of atomic symbols to add to the template, by default None.
    numbers_list  : list, optional
        List of list of atomic numbers to add to the template each sublist corresponding to one stoichiometry of interes, by default None.
    symbols_list  : list, optional
        List of list of atomic symbols to add to the template each sublist corresponding to one stoichiometry of interes, by default None.
    symbols_range : dict, optional
        Dictionary specifying ranges of composition for each element, by default None.

    print_report : bool, optional
        If True, a report is printed, by default True.

    """

    def __init__(
        self,
        template: Atoms,
        symbols: str = None,
        numbers: list = None,
        numbers_list: list = None,
        symbols_list: list = None,
        symbols_range: dict = None,
        print_report: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Both numbers and symbols cannot be specified:
        assert (
            (
                ((numbers_list is not None) is not (symbols_list is not None))
                is not (symbols_range is not None)
            )
            is not (symbols is not None)
        ) is not (numbers is not None)  # XOR

        if numbers is not None:
            self.set_numbers(numbers)
        if symbols is not None:
            self.set_numbers(symbols2numbers(symbols))
        if numbers_list is not None:
            self.set_numbers_list(numbers_list)

        elif symbols_list is not None:
            self.set_numbers_list(
                [symbols2numbers(symbols) for symbols in symbols_list]
            )

        elif symbols_range is not None:
            self.set_numbers_list(set_combination_matrix(symbols_range))

        if type(template) is Atoms:
            template = self.convert_to_candidate_object(template)
        self._template = template

        if self.confinement_cell is None:
            self.confinement_cell = self._template.get_cell()
            self.confinement_corner = np.array([0, 0, 0])

        if print_report:
            self.environment_report()

    def has_multiple_compositions(self):  # -> bool
        return len(self._numbers_list) > 1

    def get_template(self):
        return self._template.copy()

    def set_template(self, template):
        self._template = template

    def set_numbers(self, numbers):
        # Will override all of the _numbers list
        self._numbers_list = [numbers]
        self._numbers = self._numbers_list[0]

    def get_numbers(self):
        id = np.random.randint(len(self._numbers_list))
        return np.array(self._numbers_list[id]).flatten()

    def get_missing_types(self):
        return np.sort(np.unique(self.get_numbers()))

    def get_all_types(self):  # -> listlist[int]:
        return list(set(list(self._template.numbers) + self.get_numbers_list_flat()))

    def get_identifier(self):
        return self.__hash__()

    def get_missing_indices(self):
        max_additional = max([len(n) for n in self._numbers_list])
        return np.arange(len(self._template), len(self._template) + max_additional)

    def get_numbers_by_index(self, idx):
        return self._numbers_list[idx]

    def set_numbers_list(self, numbers_list):  # list[list[int]]):
        self._numbers_list = numbers_list
        self._numbers = [
            self._numbers_list[
                max(
                    range(len(self._numbers_list)),
                    key=lambda x: len(self._numbers_list[x]),
                )
            ]
        ]

    def get_numbers_list(self):  # -> listlist[list[int]]:
        return self._numbers_list

    def get_numbers_list_flat(self):  # -> listlist[int]:
        fl = []
        for nl in self.get_numbers_list():
            fl.extend(nl)
        return fl

    # Returns types of one or all the systems including template
    def get_types(self):  # -> listlist[int]:
        return list(set(list(self._template.numbers) + self.get_numbers()))

    # Returns types of one or all the systems including template
    def get_all_types_notemplate(self):  # ->list[int]:
        return list(set(list(self.get_numbers_list_flat())))

    def get_numbers_indices(self):
        return np.arange(len(self._template), len(self._template) + len(self._numbers))

    def get_all_numbers(self):
        all_numbers = np.append(self.get_numbers(), self._template.get_atomic_numbers())
        return all_numbers

    def get_numbers_types_by_index(self, idx):
        return np.sort(np.unique(self._numbers_list[idx].copy()))

    def get_numbers_indices_by_index(self, idx):
        return np.arange(
            len(self._template),
            len(self._template) + len(self.get_numbers_types_by_index(idx)),
        )

    def get_full_formula(self):
        return np.append(self.get_numbers(), self._template.get_atomic_numbers())

    def get_full_formula_by_index(self, idx):
        return np.append(
            self.get_numbers_by_index(idx), self._template.get_atomic_numbers()
        )

    def get_all_species(self):
        return list(Symbols(self.get_all_types()).species())

    def get_species(self):
        return list(np.unique(self.get_all_species()))

    def get_atoms(self):
        atoms = self.get_template()
        atoms += Atoms(self.get_numbers())
        return atoms

    def get_atoms_by_index(self, idx):  # -> Atoms
        atoms = self.get_template()
        atoms += Atoms(self.get_numbers_list()[idx % len(self.get_numbers_list())])
        return atoms

    def get_all_atoms(self):
        n_list = self.get_numbers_list()
        return [self.get_template() + Atoms([at]) for at in n_list]

    def match(self, candidate):
        cand_numbers = candidate.get_atomic_numbers()
        if not (
            candidate.positions[0 : len(candidate.template)] == self._template.positions
        ).all():
            return False
        c = cand_numbers[len(candidate.template) :]
        for i, n in enumerate(self._numbers_list):
            if (np.sort(c) == np.sort(n)).all():
                return True
        return False

    def get_candidate_numbers_strict(self, candidate):
        """
        Returns the list with the numbers specific for that candidate
        Checks if candidate has same template"""
        assert [
            candidate[idx].symbol for idx in candidate.get_template_indices()
        ] == list(self._template.symbols)
        numbers = sorted(
            [candidate[i].number for i in candidate.get_optimize_indices()]
        )
        for idx, n in enumerate(self._numbers_list):
            if numbers == sorted(n):
                return n, idx

        return None, -1

    def get_candidate_numbers(self, candidate):
        """
        Returns the list with the numbers specific for that candidate
        Does not check if candidate has the same tempalte"""
        numbers = sorted([candidate[i] for i in candidate.get_optimize_indices()])
        for n in self._numbers_list:
            if (numbers == sorted(n)).all():
                return n
        return None

    def __hash__(self):
        feature = (
            tuple(self.get_numbers())
            + tuple(self._template.get_atomic_numbers())
            + tuple(self._template.get_positions().flatten().tolist())
        )
        return hash(feature)

    def get_identifier(self):
        return self.__hash__()

    def environment_report(self) -> None:
        self.writer.write_panel(str(self), "Environment report")

    def __str__(self) -> str:
        string = ""
        tab = "    "
        string += "Atoms in search:\n"

        missing_numbers = self.get_numbers()
        for number in np.unique(missing_numbers):
            symbols_object = Symbols([number])
            specie = symbols_object.species().pop()
            count = np.count_nonzero(missing_numbers == number)
            string += f"{tab}{specie} = {count}\n"

        total_symbols = Symbols(self.get_all_numbers())
        string += f"Template formula: {self._template.get_chemical_formula()}\n"
        string += f"Full formula: {total_symbols.get_chemical_formula()}\n"

        string += "Cell:\n"
        for cell_vec in self._template.get_cell():
            string += tab + "{:4.2f} {:4.2f} {:4.2f}\n".format(*cell_vec)
        
        string += "Periodicity:\n"
        string += tab + "{} {} {}\n".format(*self._template.pbc)

        string += f"Box constraint: {self.use_box_constraint}\n"
        
        if self.use_box_constraint:
            assert self.confinement_cell is not None
            assert self.confinement_corner is not None
            string += "Confinement corner\n"
            string += tab + "{:4.2f} {:4.2f} {:4.2f}\n".format(*self.confinement_corner)
            string += "Confinement cell:\n"
            for i, cell_vec in enumerate(self.confinement_cell):
                string += tab + "{:4.2f} {:4.2f} {:4.2f}".format(*cell_vec)
                if i != 2:
                    string += "\n"
        return string
        
